// URL: http://lz4.googlecode.com/svn/trunk
// Revision: 88
//
// Note: This revision seems to cause some bad memory accesses when using LZ4_compressHC.
//       See our wrapper Lz4Compression.cpp for details.