#ifndef tuvok_internal_mangle_h8
#define tuvok_internal_mangle_h8

/*
 * This header file mangles all symbols exported from the jpeg library.
 * It is included in all files while building the jpeg library.
 *
 * This idea/implementation was taken from GDCM.
 */

#define jcopy_block_row tuvok_internal_jcopy_block_row8
#define jcopy_sample_rows tuvok_internal_jcopy_sample_rows8
#define jdiv_round_up tuvok_internal_jdiv_round_up8
#define jinit_1pass_quantizer tuvok_internal_jinit_1pass_quantizer8
#define jinit_2pass_quantizer tuvok_internal_jinit_2pass_quantizer8
#define jinit_arith_decoder tuvok_internal_jinit_arith_decoder8
#define jinit_arith_encoder tuvok_internal_jinit_arith_encoder8
#define jinit_c_codec tuvok_internal_jinit_c_codec8
#define jinit_c_coef_controller tuvok_internal_jinit_c_coef_controller8
#define jinit_c_diff_controller tuvok_internal_jinit_c_diff_controller8
#define jinit_c_main_controller tuvok_internal_jinit_c_main_controller8
#define jinit_c_master_control tuvok_internal_jinit_c_master_control8
#define jinit_c_prep_controller tuvok_internal_jinit_c_prep_controller8
#define jinit_c_scaler tuvok_internal_jinit_c_scaler8
#define jinit_color_converter tuvok_internal_jinit_color_converter8
#define jinit_color_deconverter tuvok_internal_jinit_color_deconverter8
#define jinit_compress_master tuvok_internal_jinit_compress_master8
#define jinit_d_codec tuvok_internal_jinit_d_codec8
#define jinit_d_coef_controller tuvok_internal_jinit_d_coef_controller8
#define jinit_d_diff_controller tuvok_internal_jinit_d_diff_controller8
#define jinit_d_main_controller tuvok_internal_jinit_d_main_controller8
#define jinit_d_post_controller tuvok_internal_jinit_d_post_controller8
#define jinit_d_scaler tuvok_internal_jinit_d_scaler8
#define jinit_differencer tuvok_internal_jinit_differencer8
#define jinit_downsampler tuvok_internal_jinit_downsampler8
#define jinit_forward_dct tuvok_internal_jinit_forward_dct8
#define jinit_input_controller tuvok_internal_jinit_input_controller8
#define jinit_inverse_dct tuvok_internal_jinit_inverse_dct8
#define jinit_lhuff_decoder tuvok_internal_jinit_lhuff_decoder8
#define jinit_lhuff_encoder tuvok_internal_jinit_lhuff_encoder8
#define jinit_lossless_c_codec tuvok_internal_jinit_lossless_c_codec8
#define jinit_lossless_d_codec tuvok_internal_jinit_lossless_d_codec8
#define jinit_lossy_c_codec tuvok_internal_jinit_lossy_c_codec8
#define jinit_lossy_d_codec tuvok_internal_jinit_lossy_d_codec8
#define jinit_marker_reader tuvok_internal_jinit_marker_reader8
#define jinit_marker_writer tuvok_internal_jinit_marker_writer8
#define jinit_master_decompress tuvok_internal_jinit_master_decompress8
#define jinit_memory_mgr tuvok_internal_jinit_memory_mgr8
#define jinit_merged_upsampler tuvok_internal_jinit_merged_upsampler8
#define jinit_phuff_decoder tuvok_internal_jinit_phuff_decoder8
#define jinit_phuff_encoder tuvok_internal_jinit_phuff_encoder8
#define jinit_shuff_decoder tuvok_internal_jinit_shuff_decoder8
#define jinit_shuff_encoder tuvok_internal_jinit_shuff_encoder8
#define jinit_undifferencer tuvok_internal_jinit_undifferencer8
#define jinit_upsampler tuvok_internal_jinit_upsampler8
#define jpeg_CreateCompress tuvok_internal_jpeg_CreateCompress8
#define jpeg_CreateDecompress tuvok_internal_jpeg_CreateDecompress8
#define jpeg_abort tuvok_internal_jpeg_abort8
#define jpeg_abort_compress tuvok_internal_jpeg_abort_compress8
#define jpeg_abort_decompress tuvok_internal_jpeg_abort_decompress8
#define jpeg_add_quant_table tuvok_internal_jpeg_add_quant_table8
#define jpeg_alloc_huff_table tuvok_internal_jpeg_alloc_huff_table8
#define jpeg_alloc_quant_table tuvok_internal_jpeg_alloc_quant_table8
#define jpeg_calc_output_dimensions tuvok_internal_jpeg_calc_output_dimensions8
#define jpeg_consume_input tuvok_internal_jpeg_consume_input8
#define jpeg_copy_critical_parameters tuvok_internal_jpeg_copy_critical_parameters8
#define jpeg_default_colorspace tuvok_internal_jpeg_default_colorspace8
#define jpeg_destroy tuvok_internal_jpeg_destroy8
#define jpeg_destroy_compress tuvok_internal_jpeg_destroy_compress8
#define jpeg_destroy_decompress tuvok_internal_jpeg_destroy_decompress8
#define jpeg_fdct_float tuvok_internal_jpeg_fdct_float8
#define jpeg_fdct_ifast tuvok_internal_jpeg_fdct_ifast8
#define jpeg_fdct_islow tuvok_internal_jpeg_fdct_islow8
#define jpeg_fill_bit_buffer tuvok_internal_jpeg_fill_bit_buffer8
#define jpeg_finish_compress tuvok_internal_jpeg_finish_compress8
#define jpeg_finish_decompress tuvok_internal_jpeg_finish_decompress8
#define jpeg_finish_output tuvok_internal_jpeg_finish_output8
#define jpeg_free_large tuvok_internal_jpeg_free_large8
#define jpeg_free_small tuvok_internal_jpeg_free_small8
#define jpeg_gen_optimal_table tuvok_internal_jpeg_gen_optimal_table8
#define jpeg_get_large tuvok_internal_jpeg_get_large8
#define jpeg_get_small tuvok_internal_jpeg_get_small8
#define jpeg_has_multiple_scans tuvok_internal_jpeg_has_multiple_scans8
#define jpeg_huff_decode tuvok_internal_jpeg_huff_decode8
#define jpeg_idct_1x1 tuvok_internal_jpeg_idct_1x18
#define jpeg_idct_2x2 tuvok_internal_jpeg_idct_2x28
#define jpeg_idct_4x4 tuvok_internal_jpeg_idct_4x48
#define jpeg_idct_float tuvok_internal_jpeg_idct_float8
#define jpeg_idct_ifast tuvok_internal_jpeg_idct_ifast8
#define jpeg_idct_islow tuvok_internal_jpeg_idct_islow8
#define jpeg_input_complete tuvok_internal_jpeg_input_complete8
#define jpeg_make_c_derived_tbl tuvok_internal_jpeg_make_c_derived_tbl8
#define jpeg_make_d_derived_tbl tuvok_internal_jpeg_make_d_derived_tbl8
#define jpeg_mem_available tuvok_internal_jpeg_mem_available8
#define jpeg_mem_init tuvok_internal_jpeg_mem_init8
#define jpeg_mem_term tuvok_internal_jpeg_mem_term8
#define jpeg_natural_order tuvok_internal_jpeg_natural_order8
#define jpeg_new_colormap tuvok_internal_jpeg_new_colormap8
#define jpeg_open_backing_store tuvok_internal_jpeg_open_backing_store8
#define jpeg_quality_scaling tuvok_internal_jpeg_quality_scaling8
#define jpeg_read_coefficients tuvok_internal_jpeg_read_coefficients8
#define jpeg_read_header tuvok_internal_jpeg_read_header8
#define jpeg_read_raw_data tuvok_internal_jpeg_read_raw_data8
#define jpeg_read_scanlines tuvok_internal_jpeg_read_scanlines8
#define jpeg_resync_to_restart tuvok_internal_jpeg_resync_to_restart8
#define jpeg_save_markers tuvok_internal_jpeg_save_markers8
#define jpeg_set_colorspace tuvok_internal_jpeg_set_colorspace8
#define jpeg_set_defaults tuvok_internal_jpeg_set_defaults8
#define jpeg_set_linear_quality tuvok_internal_jpeg_set_linear_quality8
#define jpeg_set_marker_processor tuvok_internal_jpeg_set_marker_processor8
#define jpeg_set_quality tuvok_internal_jpeg_set_quality8
#define jpeg_simple_lossless tuvok_internal_jpeg_simple_lossless8
#define jpeg_simple_progression tuvok_internal_jpeg_simple_progression8
#define jpeg_start_compress tuvok_internal_jpeg_start_compress8
#define jpeg_start_decompress tuvok_internal_jpeg_start_decompress8
#define jpeg_start_output tuvok_internal_jpeg_start_output8
#define jpeg_std_error tuvok_internal_jpeg_std_error8
#define jpeg_std_message_table tuvok_internal_jpeg_std_message_table8
#define jpeg_stdio_dest tuvok_internal_jpeg_stdio_dest8
#define jpeg_stdio_src tuvok_internal_jpeg_stdio_src8
#define jpeg_suppress_tables tuvok_internal_jpeg_suppress_tables8
#define jpeg_write_coefficients tuvok_internal_jpeg_write_coefficients8
#define jpeg_write_m_byte tuvok_internal_jpeg_write_m_byte8
#define jpeg_write_m_header tuvok_internal_jpeg_write_m_header8
#define jpeg_write_marker tuvok_internal_jpeg_write_marker8
#define jpeg_write_raw_data tuvok_internal_jpeg_write_raw_data8
#define jpeg_write_scanlines tuvok_internal_jpeg_write_scanlines8
#define jpeg_write_tables tuvok_internal_jpeg_write_tables8
#define jround_up tuvok_internal_jround_up8
#define jzero_far tuvok_internal_jzero_far8

#define first_marker tuvok_internal_first_marker8
#define read_markers tuvok_internal_read_markers8
#define consume_markers tuvok_internal_consume_markers8

#endif // tuvok_internal_mangle_h8
