/*
 For more information, please see: http://software.sci.utah.edu

 The MIT License

 Copyright (c) 2012 Scientific Computing and Imaging Institute,
 University of Utah.


 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.
 */

/**
 \brief Subclasses QMdiSubWindow so that we have more complete control over
        the destruction of a render window. Necessary so that undo functions
        properly when the window is destroyed.
 */

#include "../Tuvok/StdTuvokDefines.h"
#include <cassert>
#include <sstream>
#include <stdexcept>
#include "GL/glew.h"
#if defined(__GNUC__) && defined(DETECTED_OS_LINUX)
# pragma GCC visibility push(default)
#endif
#include <QtGui/QtGui>
#if defined(__GNUC__) && defined(DETECTED_OS_LINUX)
# pragma GCC visibility pop
#endif

#include "MDIRenderWin.h"
#include "RenderWindow.h"

using namespace tuvok;

MDIRenderWin::MDIRenderWin(MasterController& masterController,
                           RenderWindow* renderWin,
                           QWidget *parent,
                           Qt::WindowFlags flags)
: QMdiSubWindow(parent, flags)
, mMasterController(masterController)
, mRenderWindow(renderWin)
{
  setWidget(renderWin->GetQtWidget());
}

MDIRenderWin::~MDIRenderWin()
{

}


void MDIRenderWin::closeEvent(QCloseEvent* event)
{
  QMdiSubWindow::closeEvent(event);

  // Kill off our child widget (in this case, the render window that was
  // given to us).
  mMasterController.LuaScript()->cexec("deleteClass",
                                       mRenderWindow->GetLuaInstance());
}


